package org.netkernel.ext.system.accessor;

import java.io.IOException;

import java.util.*;

import org.netkernel.container.ISpaceListener;
import org.netkernel.ext.system.representation.IRepDeployedModules;
import org.netkernel.ext.system.representation.impl.ProtoMap;
import org.netkernel.ext.system.representation.impl.RepSingleModule;
import org.netkernel.ext.system.representation.impl.RepAllModules;
import org.netkernel.ext.system.representation.impl.RepresentationReferences;
import org.netkernel.ext.system.representation.impl.SpaceGraph;
//import org.netkernel.ext.system.representation.impl.SpaceGraph;
import org.netkernel.ext.system.representation.impl.RepSingleModule.HashedItem;
import org.netkernel.layer0.boot.BootOrderOptimizer;
import org.netkernel.layer0.boot.IModule;
import org.netkernel.layer0.boot.ModuleManager;
import org.netkernel.layer0.boot.ModuleSourceWithLevel;
import org.netkernel.layer0.meta.*;
import org.netkernel.layer0.meta.impl.DefaultModuleMeta;
import org.netkernel.layer0.meta.impl.EndpointMetaBuilder;
import org.netkernel.layer0.meta.impl.EndpointMetaImpl;
import org.netkernel.layer0.meta.impl.PhysicalEndpointMetaImpl;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.nkf.impl.NKFResponseImpl;
import org.netkernel.layer0.nkf.impl.NKFResponseReadOnlyImpl;
import org.netkernel.layer0.nkf.impl.NKFTransportContextImpl;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.representation.impl.HDSNodeImpl;
import org.netkernel.layer0.urii.AggregatingPolledExpiry;
import org.netkernel.layer0.urii.ParsedIdentifierImpl;
import org.netkernel.layer0.util.Utils;
import org.netkernel.module.standard.endpoint.IStandardEndpoint;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.module.standard.endpoint.StandardEndpointImpl;
import org.netkernel.request.IExpiry;
import org.netkernel.request.IRequestResponseFields;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.urii.*;

public class ModuleListAccessor extends StandardAccessorImpl implements ISpaceListener
{
	private Map mModuleToHash;
	private Map mHashToModule;
	
	private INKFResponseReadOnly<IRepDeployedModules> mLastWholeResponse;
	
	private long mLastSync;
	
	private final IMetaRepresentation mBrokenMeta;
	
	private long mHysteresis;
	
	public ModuleListAccessor()
	{	this.declareThreadSafe();
		this.declareName("Module List");
		this.declareDescription("lists all modules and their contexts");
		this.declareSourceRepresentation(IRepDeployedModules.class);
		this.declareSourceRepresentation(IReadableBinaryStreamRepresentation.class);
		EndpointMetaImpl m=new EndpointMetaImpl("[broken metadata]", null, null, null, 0, EndpointMetaBuilder.DEFAULT_GRAMMAR, EndpointMetaImpl.NO_ARGS, EndpointMetaImpl.NO_RESPONSES, EndpointMetaImpl.NO_EXCEPTIONS);
		m.addAdditionalField(IEndpointStateMeta.KEY_PROBLEM, true);
		mBrokenMeta=m;
	}

	public void spaceChanged()
	{	mLastWholeResponse=null;
		
		final NKFTransportContextImpl c=new NKFTransportContextImpl(getKernel(), getSpace(), this);
		Thread t = new Thread()
		{
			public void run()
			{	try
				{	Thread.sleep(mHysteresis);
					c.source("active:moduleList");
					c.logRaw(INKFLocale.LEVEL_INFO, "ModuleListAccessor complete");
				}
				catch (Exception e)
				{
				}
			}
		};
		t.start();
	}	
	
	public void postCommission(INKFRequestContext aContext) throws Exception
	{	mHysteresis=getKernel().getConfiguration().getLong("stdmod.initialisation.hysteresis", 500L);
		getKernel().addSpaceListener(this);
	}
	
	public void preDecommission(INKFRequestContext aContext) throws Exception
	{	getKernel().removeSpaceListener(this);
	}
	
	public void onSource(INKFRequestContext aContext) throws Exception
	{
		boolean part = aContext.getThisRequest().argumentExists("moduleHash");
		if (part)
		{	String moduleHash = aContext.getThisRequest().getArgumentValue("moduleHash");
			generateSingleModuleRep(moduleHash,aContext);
		}
		else
		{	String type=aContext.getThisRequest().getArgumentValue(ParsedIdentifierImpl.ARG_ACTIVE_TYPE);
			if (type.equals("moduleList"))
			{	IRepDeployedModules rep=getWhole(aContext);
				aContext.createResponseFrom(rep);
			}
			else if (type.equals("moduleListWhole"))
			{	generateWhole(aContext);
			}
			else if (type.equals("moduleListDoc"))
			{	IRepDeployedModules rep=getWhole(aContext);
				IHDSNode hierarchy = rep.getHierarchy();
				aContext.createResponseFrom(hierarchy);
			}
			else if (type.equals("moduleListProtoMap"))
			{	IRepDeployedModules rep=getWhole(aContext);
				ProtoMap result = new ProtoMap(rep);
				aContext.createResponseFrom(result);
			}
			else if (type.equals("moduleListSpaceGraph"))
			{	IRepDeployedModules rep=getWhole(aContext);
				SpaceGraph result=new SpaceGraph(rep);
				aContext.createResponseFrom(result);
			}
			else if (type.equals("moduleListRepMap"))
			{	IRepDeployedModules rep=getWhole(aContext);
				RepresentationReferences result = new RepresentationReferences(rep);
				aContext.createResponseFrom(result);
			}
		}
	}
	
	private synchronized IRepDeployedModules getWhole( INKFRequestContext aContext) throws Exception
	{	if (mLastWholeResponse==null || mLastWholeResponse.isExpired())
		{	mLastWholeResponse=aContext.sourceForResponse("active:moduleListWhole",IRepDeployedModules.class);
		}
		else
		{	aContext.getKernelContext().declareDependency(mLastWholeResponse);
		}
		return mLastWholeResponse.getRepresentation();
	}
	
	private void generateWhole(INKFRequestContext aContext) throws Exception
	{
		BootOrderOptimizer boo=BootOrderOptimizer.sInstance;
		
		Map<String,Map<String, Object>> itemMaps = new HashMap(64);
		Map<Object,String> refToHash = new IdentityHashMap(4096);
		Map<String,String> elementToHash = new HashMap(1024);

		HDSNodeImpl root = new HDSNodeImpl(null,null);
		HDSNodeImpl root2 = new HDSNodeImpl("modules",root);
		root.setChildren(new IHDSNode[]{root2});
		List<IHDSNode> moduleNodes = new ArrayList<IHDSNode>();
		
		ModuleManager mm=ModuleManager.getSingleton();
		List<IModule> modules = mm.getModules();
		boo.sort(modules);
		
		boo.reset(aContext.getKernelContext().getKernel());
		
		List<INKFAsyncRequestHandle> handles=new ArrayList<INKFAsyncRequestHandle>(modules.size());
		List<IExpiry> expiries=new ArrayList<IExpiry>(modules.size());
		for (Iterator i=modules.iterator(); i.hasNext(); )
        {	IModule module = (IModule)i.next();
        	String moduleHash = getHashForModule(module);
               
            INKFRequest req = aContext.createRequest("active:moduleList");
            req.addArgument("moduleHash", moduleHash);
            req.setRepresentationClass(RepSingleModule.class);
            INKFResponseReadOnly<RepSingleModule> respIn=aContext.issueRequestForResponse(req);
            RepSingleModule rep = (RepSingleModule)respIn.getRepresentation();
            expiries.add(((NKFResponseReadOnlyImpl)respIn).getMeta().getExpiry());
            
            Map<String, Object> items=rep.getItems();
            itemMaps.put(rep.getHash(),items);
            mergeMap(refToHash,rep.getRefs());
            mergeMap(elementToHash,rep.getElements());

            for (IHDSNode n: rep.getHierarchy().getChildren())
            {    moduleNodes.add(n);
            }
        } 
		
		root2.setChildren(moduleNodes);
		
		IRepDeployedModules result = new RepAllModules(root,itemMaps,refToHash,elementToHash);
		
		INKFResponse resp=aContext.createResponseFrom(result);
		long expiryPeriod=getKernel().getConfiguration().getLong("netkernel.poll.internal", 500L);
		resp.setExpiry(INKFResponse.EXPIRY_FUNCTION, new AggregatingPolledExpiry(expiries, expiryPeriod, "active:moduleList"));
		resp.setMimeType("text/xml");
		
		boo.analyze(modules,aContext);
		boo.store(aContext);
		
		
	}
	
	private void mergeMap(Map aBase, Map aAdditions)
	{	for (Iterator i=aAdditions.entrySet().iterator(); i.hasNext(); )
		{	Map.Entry entry = (Map.Entry)i.next();
			aBase.put(entry.getKey(), entry.getValue());
		}
	}
		
	private void outputMeta(HDSBuilder b, IMetaRepresentation aMeta) //throws IOException
	{
		String name = aMeta.getName();
		if (name!=null)
		{	b.addNode("name", name);
		}
		String desc = aMeta.getDescription();
		if (desc!=null)
		{	b.addNode("desc", desc);
		}
		IIdentifier icon = aMeta.getIcon();
		if (icon!=null)
		{	b.addNode("icon", icon);
		}
		IIdentifier hrd = aMeta.getHumanReadableDocumentation();
		if (hrd!=null)
		{	b.addNode("hrd", hrd);
		}				
	}	
	
	private void outputMeta(HDSBuilder b, IModuleMeta aMeta) //throws IOException
	{	String source = aMeta.getSource().toString();
		if (source!=null)
		{	b.addNode("source", source);
		}
		String version = aMeta.getVersion();
		if (version!=null)
		{	b.addNode("version", version);
		}
		outputMeta(b, (IMetaRepresentation)aMeta);
	}
	
	
	private void generateSingleModuleRep(String aModuleHash, INKFRequestContext aContext) throws IOException
	{
		HDSBuilder b = new HDSBuilder();
		String hashBase=aModuleHash+"/";
		MLState2 state2 = new MLState2(aContext,hashBase);
		IModule module=getModuleForHash(aModuleHash);
		
		b.pushNode("module");		
		b.addNode("hash", aModuleHash);
		ModuleSourceWithLevel mswl = ModuleManager.getSingleton().getRegisteredModules().get(module.getSource());
		if (mswl!=null)
		{	b.addNode("runlevel", mswl.getLevel());
		}
		
		IModuleMeta meta=module.getMeta();
		if (meta==null)
		{	meta=new DefaultModuleMeta(module);
		}
		outputMeta(b, meta);
		b.addNode("version", meta.getVersion());
		b.addNode("id", meta.getIdentifier());
		
		HashedItem hi = new HashedItem(module,meta);
		state2.mHashedItems.put(aModuleHash, hi);
		state2.mHashForRef.put(module, aModuleHash);
		state2.mHashForRef.put(meta, aModuleHash);
		
		b.pushNode("spaces");
		
		ISpace[] spaces=module.getSpaces();
		Set<ISpace> usedSpaces=new HashSet<ISpace>();
		
		boolean moduleProblems=false;
		for (int j=0; j<spaces.length; j++)
		{	ISpace space=spaces[j];
			state2.mHashBase=hashBase;
			moduleProblems |=writeSpace(space,b,state2, usedSpaces);
		}
		b.popNode();
		
		if (state2.mPublicSpace!=null)
		{	b.addNode("publicSpaceHash", state2.mPublicSpace);
		}
		
		if (!moduleProblems)
		{	IRequestResponseFields moduleFields = meta.getAdditionalFields();
			moduleProblems|=(moduleFields.getValue(IEndpointStateMeta.KEY_PROBLEM)!=null);
		}
		if (moduleProblems)
		{	b.addNode(IEndpointStateMeta.KEY_PROBLEM, null);
		}
		
		b.popNode();
		
		RepSingleModule result = new RepSingleModule(aModuleHash,b.getRoot(),state2.mHashedItems,state2.mHashForRef, state2.mHashForElement);
		INKFResponse resp=aContext.createResponseFrom(result);
		//stop fast invalidation that can cause system to grind to a halt
		resp.setExpiry(INKFResponse.EXPIRY_MAX_CONSTANT_DEPENDENT,System.currentTimeMillis()+mHysteresis);
		resp.setHeader("visualizer_ignore", Boolean.TRUE);
		resp.setMimeType("text/xml");
		//resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
	}
	
	private static class MLState2
	{
		public Map<String, Object> mHashedItems = new HashMap<String, Object>(128);
		public Map<Object, String> mHashForRef = new IdentityHashMap<Object, String>(128);
		public Map<String, String> mHashForElement = new HashMap<String, String>(64);
		public INKFRequestContext mContext;
		public String mPublicSpace;
		public String mHashBase;
		private int mSpaceIndex;
		
		public MLState2(INKFRequestContext aContext,String aHashBase)
		{	mContext=aContext;
			mHashBase=aHashBase;
			mSpaceIndex=mHashBase.hashCode();
		}
		
		public int getNextSpaceIndex()
		{	return mSpaceIndex++;
		}
		
	}
	
	private boolean writeSpace(ISpace aSpace, HDSBuilder b, MLState2 aState2, Set<ISpace> aUsedSpaces)
	{
		if (aUsedSpaces.contains(aSpace)) return false;
		aUsedSpaces.add(aSpace);
		//determine identity
		INKFRequestContext context=aState2.mContext;
		
		//determine space meta
		ISpaceMeta spaceMeta;
		try
		{	INKFRequest req = context.createRequest( "" );
			req.setVerb(INKFRequestReadOnly.VERB_META);
			IRequestScopeLevel parentRctx=context.getKernelContext().getThisKernelRequest().getRequestScope();
			IRequestScopeLevel rctx = RequestScopeLevelImpl.createOrphanedRootScopeLevel(aSpace,parentRctx);
			req.setRequestScope( rctx );
			req.setRepresentationClass( ISpaceMeta.class );
			spaceMeta = (ISpaceMeta)context.issueRequest(req);	
		}
		catch (Exception e)
		{	context.logFormatted(INKFLocale.LEVEL_WARNING, "EX_GET_META_FAILED", aSpace, Utils.throwableToString(e));
			return true;
		}
		
		//setup hashing for space
		b.pushNode("space");
		HashedItem hi = new HashedItem(aSpace,spaceMeta);
		String hashString = putUniqueHash(aState2.mHashBase,aState2.getNextSpaceIndex(), 1, hi, aState2.mHashedItems);
		aState2.mHashForRef.put(aSpace, hashString);
		aState2.mHashForRef.put(spaceMeta, hashString);
		
		boolean spaceProblems=false;
		boolean moduleProblems=false;
		b.addNode("hash", hashString);
		String identifier="";
		boolean isPublic=false;
		if (aSpace instanceof ISpaceWithIdentity)
		{	ISpaceWithIdentity spi=(ISpaceWithIdentity)aSpace;
			if (spi.getIdentifier()!=null)
			{	identifier=spi.getIdentifier().toString();
				b.addNode("id", identifier);
				b.addNode("version", spi.getVersion());
				
				ISpaceWithIdentity spaceFromKernel=getKernel().getSpace(spi.getIdentifier(), spi.getVersion(), spi.getVersion());
				isPublic=(spaceFromKernel!=null);
			}
		}		
		b.addNode("public",isPublic);
		outputMeta(b,spaceMeta);
		
		if (isPublic && aState2.mPublicSpace==null)
		{	//public space is assign to first public space exposed by module
			aState2.mPublicSpace = hashString;
		}
		
		aState2.mHashBase=hashString;
		boolean[] problemArray = writeMappings(aSpace,spaceMeta,b,aState2, aUsedSpaces);
		spaceProblems|=problemArray[0];
		moduleProblems|=problemArray[1];
		
		if (!spaceProblems)
		{	IRequestResponseFields spaceFields=spaceMeta.getAdditionalFields();
			spaceProblems|=(spaceFields.getValue(IEndpointStateMeta.KEY_PROBLEM)!=null);
		}
		if (spaceProblems)
		{	b.addNode(IEndpointStateMeta.KEY_PROBLEM, null);
		}			
		
		b.popNode();
		return spaceProblems||moduleProblems;
	}
	
	private boolean[] writeMappings(ISpace aSpace, ISpaceMeta aSpaceMeta, HDSBuilder b, MLState2 aState2, Set<ISpace> aUsedSpaces)
	{
		INKFRequestContext context=aState2.mContext;
		ISpaceElements elements=aSpaceMeta.getElements();
		int s=elements.size();
		
		//output mappings
		boolean problems=false;
		//b.pushNode("mappings");				
		
		List<IEndpoint> physicalEndpoints = new ArrayList<IEndpoint>();
		
		String elementHashBase=aState2.mHashBase+"/";
		IRequestScopeLevel parentRctx=context.getKernelContext().getThisKernelRequest().getRequestScope();
		IRequestScopeLevel rctx = RequestScopeLevelImpl.createOrphanedRootScopeLevel(aSpace,parentRctx);

		for (int l=0; l<s; l++)
		{	IEndpoint physicalEndpoint=elements.getPhysicalEndpoint(l);
			if (physicalEndpoint!=null && !physicalEndpoints.contains(physicalEndpoint))
			{	physicalEndpoints.add(physicalEndpoint);				

				//b.pushNode("mapping");
				IPhysicalEndpointMeta peMeta=new PhysicalEndpointMetaImpl(physicalEndpoint);
				HashedItem hi=new HashedItem(physicalEndpoint,peMeta);
				String epHashString = putUniqueHash(elementHashBase,l, 2, hi, aState2.mHashedItems);
				aState2.mHashForRef.put(physicalEndpoint, epHashString);
				aState2.mHashForRef.put(peMeta, epHashString);
				
				boolean mappingProblems=false;
				
				try
				{	IRequestResponseFields elementFields=peMeta.getAdditionalFields();
					if (elementFields.getValue(IEndpointStateMeta.KEY_PROBLEM)!=null)
					{	mappingProblems=true;
					}
				}
				catch (Exception e)
				{	
				}
				
				//b.addNode("hash", epHashString);
				//outputMeta(b, peMeta);
				
				//b.pushNode("elements");
				for (int i=0; i<s; i++)
				{	if (elements.getPhysicalEndpoint(i)==physicalEndpoint)
					{	IIdentifier elementId=elements.getIdentifier(i);
						if (elementId!=null)
						{
							//b.pushNode("element");
							String idString = elementId.toString();
							//b.addNode("id", idString);

							String elementHash=elementHashBase+idString;
							try
							{	INKFRequest req = context.createRequest( idString );
								req.setVerb(INKFRequestReadOnly.VERB_META);
								req.setRequestScope( rctx );
								req.setRepresentationClass( IMetaRepresentation.class );
								IMetaRepresentation elementMeta = (IMetaRepresentation)context.issueRequest(req);
								
								
								//hi=new HashedItem(null,elementMeta);
								aState2.mHashedItems.put(elementHash, elementMeta);
								aState2.mHashForRef.put(elementMeta, elementHash);
								
								//ensure that element id is only indexed if it isn't delegated, i.e. we only index the origin
								ISpace partDelegation=elements.getDelegatedSpace(l);
								if (partDelegation==null || partDelegation==aSpace)
								{	aState2.mHashForElement.put(idString, elementHash);
									//b.addNode("primary", Boolean.TRUE);
								}
								
								//b.addNode("hash", elementHash);
								
								String type;
								if (elementMeta instanceof IEndpointMeta)
								{	type="endpoint";
								}
								else if (elementMeta instanceof IPrototypeMeta)
								{	type="prototype";
								}
								else if (elementMeta instanceof IRepresentationMeta)
								{	type="representation";
									IRepresentationMeta rm = (IRepresentationMeta)elementMeta;
									Class impClass = rm.getImplementation();
									if (partDelegation==null || partDelegation==aSpace)
									{	aState2.mHashForRef.put(impClass, elementHash);
									}
								}
								else
								{	type="";
								}
								
								if (elementMeta!=null)
								{	//b.addNode("type", type);
									
									//outputMeta(b,elementMeta);
									
									try
									{	IRequestResponseFields elementFields=elementMeta.getAdditionalFields();
										if (elementFields.getValue(IEndpointStateMeta.KEY_PROBLEM)!=null)
										{	b.addNode(IEndpointStateMeta.KEY_PROBLEM, null);
											mappingProblems=true;
										}
									}
									catch (Exception e)
									{	
									}
								}
							}
							catch (Exception e)
							{	//context.logFormatted(INKFLocale.LEVEL_WARNING, "EX_GET_ELEMENT_META_FAILED", idString, aSpace);
								problems=true;
								aState2.mHashedItems.put(elementHash, mBrokenMeta);
								if (physicalEndpoint instanceof StandardEndpointImpl)
								{ ((StandardEndpointImpl)physicalEndpoint).setInitialisationProblem("Broken Metadata");
									//System.out.println("here2 "+elementHash);
								}
							}									
							
							//b.popNode();
						}
					}
				} //loop-elements
				
				//b.popNode();

				if (mappingProblems)
				{	//b.addNode(IEndpointStateMeta.KEY_PROBLEM, null);
					problems=true;
				}
				
				//b.popNode();
				
				
			} //end-if logical endpoint
			
		} //loop-mappings
		
		//b.popNode();		
		
		//output nested spaces
		boolean nestedSpaceProblems=false;
		b.pushNode("spaces");
		int j=aState2.mHashBase.indexOf('/');
		String modHash=aState2.mHashBase.substring(0,j+1);
		for (IEndpoint pe : physicalEndpoints)
		{
			if (pe instanceof IStandardEndpoint)
			{
				IStandardEndpoint spe=(IStandardEndpoint)pe;
				try
				{	ISpace[] subspaces=spe.getSpaces(context);
					for (int i=0; i<subspaces.length; i++)
					{	ISpace subspace=subspaces[i];
						aState2.mHashBase=modHash;
						nestedSpaceProblems|=writeSpace(subspace, b,aState2, aUsedSpaces);
					}
				}
				catch(Exception e)
				{	context.logRaw(INKFLocale.LEVEL_WARNING, Utils.throwableToString(e));
					problems=true;
				}
			}
		}
		b.popNode();
		
		return new boolean[] { problems, nestedSpaceProblems };
		
	}
	
	private String getHashForModule(IModule aModule)
	{	initialise();
		return (String)mModuleToHash.get(aModule);
	}
	
	private IModule getModuleForHash(String aHash)
	{	initialise();
		return (IModule)mHashToModule.get(aHash);
	}
	
	private void initialise()
	{	ModuleManager mm=ModuleManager.getSingleton();
		long lastSync = mm.getLastSyncTime();
		if (mModuleToHash==null || lastSync!=mLastSync)
		{	
			IdentityHashMap m1 = new IdentityHashMap(64);
			Map m2 = new HashMap(64);
			
			for (Iterator i=mm.getModules().iterator(); i.hasNext(); )
			{	IModule module = (IModule)i.next();
				int moduleHash=module.getSource().toString().hashCode();
				String hashString=putUniqueHash("",moduleHash,2,module,m2);
				m1.put(module, hashString);
			}
			mModuleToHash=m1;
			mHashToModule=m2;
			
		}
	}
			
	private String putUniqueHash(String aBase, int aHashCode, int aSize, Object aValue, Map aMap)
	{	aHashCode=Utils.scatterHash(aHashCode);
		String hashString;
		while(true)
		{	hashString = aBase+hashString(aHashCode,aSize);
			Object existing=aMap.put(hashString,aValue);
			if (existing==null)
			{	break;
			}
			else
			{	aMap.put(hashString,existing);
				aHashCode++;
			}
		}
		return hashString;
	}
	
	private String hashString(int aHash, int aSize)
	{ 	if (aHash<0)
		{	aHash=-aHash;
		}
		final int N=62;
		char[] result = new char[aSize];
		for (int i=0; i<aSize; i++)
		{	int d = aHash%N;
			int c;
			if (d<26)
			{	c='A'+d;
			}
			else if (d<52)
			{	c='a'+d-26;
			}
			else
			{	c='0'+d-52;
			}
			result[i]=(char)c;
			aHash/=N;
		}
		return new String(result);
	}

}