package org.netkernel.ext.system.accessor;

import org.netkernel.ext.system.representation.IRepDeployedModules;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.meta.ISpaceMeta;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.request.IRequestResponseFields;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.IMetaRepresentation;
import org.netkernel.urii.ISpace;
import org.netkernel.urii.ISpaceWithIdentity;
import org.w3c.dom.*;

public class UserMetaAggregator extends StandardAccessorImpl
{
	public UserMetaAggregator()
	{	declareThreadSafe();
		declareSourceRepresentation(Document.class);
	}
	
	public void onSource(INKFRequestContext aContext) throws Exception
	{	IRepDeployedModules modules=aContext.source("active:moduleList",IRepDeployedModules.class);
		IHDSNodeList spaces=modules.getHierarchy().getNodes("/modules/module/spaces//space");
		Document d=XMLUtils.newDocument();
		
		boolean filterSpace=aContext.getThisRequest().argumentExists("space");
		boolean filterVersion=aContext.getThisRequest().argumentExists("version");
		if(filterSpace)
		{	IHDSNode targetHDS=null;
			ISpaceWithIdentity targetSpaceWithIdentity=null;
			for (IHDSNode space: spaces)
			{	String hash=(String)space.getFirstValue("hash");
				ISpace spaceInstance=(ISpace)modules.componentForHash(hash);
				if (spaceInstance instanceof ISpaceWithIdentity)
				{	ISpaceWithIdentity swi=(ISpaceWithIdentity)spaceInstance;
					String spaceid=swi.getIdentifier().toString();
					String spacearg=aContext.getThisRequest().getArgumentValue("space");
					if(!spaceid.equals(spacearg))
					{	continue;						
					}
					if(filterVersion)
					{	String versionarg=aContext.getThisRequest().getArgumentValue("version");
						String versionid=swi.getVersion().toString();
						if(versionid.equals(versionarg))
						{	targetHDS=space;
							break;
						}
					}
					else
					{	if(targetHDS==null)
						{	targetHDS=space;
							targetSpaceWithIdentity=swi;
						}
						else if (swi.getVersion().isGreaterOrEqualThan(targetSpaceWithIdentity.getVersion()))
						{	targetHDS=space;
							targetSpaceWithIdentity=swi;
						}
					}
				}
			}
			HDSBuilder b=new HDSBuilder();
			if(targetHDS!=null)
			{	b.importNode(targetHDS);
			}
			else
			{	b.pushNode("dummy");		//We didn't find a match so push dummy to create empty node list for next phase
			}
			spaces=b.getRoot().getNodes("/space");
		}
		Node cursor=d.appendChild(d.createElement("spaces"));
		
		for (IHDSNode space: spaces)
		{	String hash=(String)space.getFirstValue("hash");
			ISpace spaceInstance=(ISpace)modules.componentForHash(hash);
			if (spaceInstance instanceof ISpaceWithIdentity)
			{	ISpaceMeta sm=(ISpaceMeta)modules.metadataForHash(hash);
				ISpaceElements sme=sm.getElements();
				int s=sme.size();
				boolean needSpace=false;
				for (int i=0; i<s; i++)
				{	IIdentifier lei=sme.getIdentifier(i);
					if (lei!=null)
					{	String leh=hash+"/"+lei.toString();
						IMetaRepresentation em=modules.metadataForHash(leh);
						IRequestResponseFields state=em.getAdditionalFields();
						Object value=state.getValue("meta");
						if (value instanceof Element)
						{	if (!needSpace)
							{	needSpace=true;
								cursor=cursor.appendChild(d.createElement("space"));
								cursor=cursor.appendChild(d.createElement("id"));
								XMLUtils.setText((Element)cursor, ((ISpaceWithIdentity)spaceInstance).getIdentifier().toString());
								cursor=cursor.getParentNode();
								cursor=cursor.appendChild(d.createElement("version"));
								XMLUtils.setText((Element)cursor, ((ISpaceWithIdentity)spaceInstance).getVersion().toString());
								cursor=cursor.getParentNode();
								cursor=cursor.appendChild(d.createElement("elements"));
							}
							
							cursor=cursor.appendChild(d.createElement("element"));
							cursor=cursor.appendChild(d.createElement("id"));
							XMLUtils.setText((Element)cursor, lei.toString());
							cursor=cursor.getParentNode();
							cursor.appendChild(d.importNode((Element)value, true));
							cursor=cursor.getParentNode();
						}
						
					}
				}
				if (needSpace)
				{	cursor=cursor.getParentNode();
					cursor=cursor.getParentNode();
				}
			}
		}
		aContext.createResponseFrom(d);
	}
}
