/******************************************************************************
  (c) Copyright 2002-2006, 1060 Research Ltd                                   

  This Software is licensed to You, the licensee, for use under the terms of   
  the 1060 Public License v1.0. Please read and agree to the 1060 Public       
  License v1.0 [www.1060research.com/license] before using or redistributing   
  this software.                                                               

  In summary the 1060 Public license has the following conditions.             
  A. You may use the Software free of charge provided you agree to the terms   
  laid out in the 1060 Public License v1.0                                     
  B. You are only permitted to use the Software with components or applications
  that provide you with OSI Certified Open Source Code [www.opensource.org], or
  for which licensing has been approved by 1060 Research Limited.              
  You may write your own software for execution by this Software provided any  
  distribution of your software with this Software complies with terms set out 
  in section 2 of the 1060 Public License v1.0                                 
  C. You may redistribute the Software provided you comply with the terms of   
  the 1060 Public License v1.0 and that no warranty is implied or given.       
  D. If you find you are unable to comply with this license you may seek to    
  obtain an alternative license from 1060 Research Limited by contacting       
  license@1060research.com or by visiting www.1060research.com                 

  NO WARRANTY:  THIS SOFTWARE IS NOT COVERED BY ANY WARRANTY. SEE 1060 PUBLIC  
  LICENSE V1.0 FOR DETAILS                                                     

  THIS COPYRIGHT NOTICE IS *NOT* THE 1060 PUBLIC LICENSE v1.0. PLEASE READ     
  THE DISTRIBUTED 1060_Public_License.txt OR www.1060research.com/license      

  File:          $RCSfile: ExceptionSerializer.java,v $
  Version:       $Name:  $ $Revision: 1.4 $
  Last Modified: $Date: 2008/05/22 14:13:33 $
 *****************************************************************************/
package org.netkernel.ext.system.transrepresentation;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.StringWriter;
import java.io.Writer;

import org.netkernel.ext.system.representation.IRepDeployedModules;
import org.netkernel.layer0.meta.IEndpointMeta;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.representation.ByteArrayRepresentation;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.netkernel.layer0.representation.WrappedThrowable;
import org.netkernel.layer0.representation.impl.HDSFactory;
import org.netkernel.layer0.util.Utils;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.module.standard.endpoint.StandardTransreptorImpl;
import org.netkernel.request.IRequest;
import org.netkernel.scheduler.RequestFrameError;
import org.netkernel.scheduler.RequestFrameException;
import org.netkernel.scheduler.SubrequestError;
import org.netkernel.scheduler.SubrequestException;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.INetKernelThrowable;
import org.netkernel.urii.ISpace;
import org.netkernel.urii.impl.NetKernelError;
import org.netkernel.urii.impl.NetKernelException;
import org.w3c.dom.Document;

public class ExceptionSerializer extends StandardTransreptorImpl
{
	public ExceptionSerializer()
	{	this.declareThreadSafe();
		this.declareName("ExceptionSerializer");
		this.declareDescription("serializes a NetKernelException");
		this.declareFromRepresentation(WrappedThrowable.class);
		this.declareFromRepresentation(Throwable.class);
		this.declareToRepresentation(IReadableBinaryStreamRepresentation.class);
		this.declareToRepresentation(IHDSNode.class);
		this.declareToRepresentation(Document.class);
		this.declareToRepresentation(String.class);
	}
	
	public void onTransrept(INKFRequestContext context) throws Exception
	{	
		Object o=context.getThisRequest().getPrimary();
		Throwable t;
		if (o instanceof Throwable)
		{	t=(Throwable)o;
		}
		else
		{	t=((WrappedThrowable)o).getThrowable();
		}
		
		if (t instanceof NetKernelException)
		{	NetKernelException nke=(NetKernelException)t;
			if (nke.getId().contains("NKFException") && nke.getMessage()==null)
			{	t=t.getCause();
			}
		}
		IRepDeployedModules dmr=null;
		if (context.getKernelContext().getKernel().inDebug())
		{	dmr = (IRepDeployedModules)context.source("active:moduleList", IRepDeployedModules.class);
		}
		
		Object representation=null;
		Class to=context.getThisRequest().getRepresentationClass();
		if (to==String.class)
		{	Writer w = new StringWriter(1024);
			writeAll(t, w, dmr);
			representation = w.toString();
		}
		else if (to.isAssignableFrom(IHDSNode.class))
		{	ByteArrayOutputStream baos = new ByteArrayOutputStream(1024);
			Writer w = new OutputStreamWriter(baos,"UTF-8");
			writeAll(t, w, dmr);
			ByteArrayInputStream bais=new ByteArrayInputStream(baos.toByteArray());
			representation=HDSFactory.parseXML(bais);
		}
		else
		{	ByteArrayOutputStream baos = new ByteArrayOutputStream(1024);
			Writer w = new OutputStreamWriter(baos,"UTF-8");
			writeAll(t, w, dmr);
			representation = new ByteArrayRepresentation(baos,"UTF-8");
		}		
		context.createResponseFrom(representation).setMimeType("text/xml");
	}
	
	private void writeAll(Throwable aThrowable, Writer aWriter, IRepDeployedModules aDMR) throws IOException
	{	XMLUtils.writeOpenTag(aWriter, "ex");
		writeStack(aThrowable, aWriter, aDMR);
		XMLUtils.writeCloseTag(aWriter, "ex");
		aWriter.flush();
	}
	
	private void writeStack(Throwable aThrowable, Writer aWriter, IRepDeployedModules aDMR) throws IOException
	{	Throwable t=aThrowable;
		while (t!=null)
		{	int skip=writeFrame(t,aWriter,aDMR);
			t = t.getCause();
			for (int i=0; i<skip; i++)
			{	t=t.getCause();
			}
		}
	}
	
	private static int writeFrame(Throwable aThrowable, Writer aWriter, IRepDeployedModules aDMR) throws IOException
	{
		int skip=0;
		XMLUtils.writeOpenTag(aWriter, "ex");
		if (aThrowable instanceof INetKernelThrowable)
		{	INetKernelThrowable nkt = (INetKernelThrowable)aThrowable;
		
			XMLUtils.writeEscaped(aWriter, "id", nkt.getId());
			if (aThrowable.getMessage()!=null)
			{	XMLUtils.writeEscaped(aWriter,"message",aThrowable.getMessage());
			}
			if (nkt.showNativeStack())
			{	org.netkernel.util.Utils.writeNativeStack(aThrowable, aWriter, -1);
			}
			if (nkt instanceof NetKernelException)
			{	((NetKernelException)nkt).writeAdditionalFieldsAsXML(aWriter,-1);
			}
			else if (nkt instanceof NetKernelError)
			{	((NetKernelError)nkt).writeAdditionalFieldsAsXML(aWriter,-1);
			}			
			
			if (aThrowable instanceof SubrequestException)
			{	skip=writeSREx( (SubrequestException)aThrowable, aWriter, aDMR);
			}
			else if (aThrowable instanceof SubrequestError)
			{	skip=writeSREr( (SubrequestError)aThrowable, aWriter, aDMR);
			}
		}
		else
		{	XMLUtils.write(aWriter, "id", aThrowable.getClass().getName());
			if (aThrowable.getMessage()!=null)
			{	XMLUtils.writeEscaped(aWriter,"message",aThrowable.getMessage());
			}
			org.netkernel.util.Utils.writeNativeStack(aThrowable, aWriter, -1);
		}
		
		XMLUtils.writeCloseTag(aWriter, "ex");
		return skip;
	}

	private static int writeSREx(SubrequestException aSRE, Writer aWriter, IRepDeployedModules aDMR) throws IOException
	{	return innerWriteSubRequestFrame(aSRE.getRequestee(),aSRE.getRequestOut(),aSRE.getEndpointIdentifier(),aSRE.getCause(),aWriter,aDMR);
	}

	private static int writeSREr(SubrequestError aSRE, Writer aWriter, IRepDeployedModules aDMR) throws IOException
	{	return innerWriteSubRequestFrame(aSRE.getRequestee(),aSRE.getRequestOut(),aSRE.getEndpointIdentifier(),aSRE.getCause(),aWriter,aDMR);
	}
	
	private static int innerWriteSubRequestFrame(Object aRequestee, IRequest aRequestOut, IIdentifier aEndpointId, Throwable aCause, Writer aWriter, IRepDeployedModules aDMR) throws IOException
	{
		Object endpointInstance = aRequestee;
		String endpointToString=endpointInstance.toString(); //Utils.shortClassName(endpointInstance.getClass());
		String physicalEndpointHash=(aDMR==null)?null:aDMR.hashForComponent(endpointInstance);
		String contextHash;
		String contextToString;
		if (physicalEndpointHash!=null)
		{	int i=physicalEndpointHash.lastIndexOf('/');
			contextHash=physicalEndpointHash.substring(0,i);
			if (contextHash.indexOf('/')==-1)
			{	contextHash=physicalEndpointHash;
			}
			Object o=aDMR.componentForHash(contextHash);
			ISpace ctx=(ISpace)o;
			contextToString=ctx.toString();
		}
		else
		{	ISpace ctx = aRequestOut.getRequestScope().getSpace();
			contextHash=(aDMR==null)?null:aDMR.hashForComponent(ctx);
			contextToString=ctx.toString();
		}
		
		if (contextHash!=null)
		{	XMLUtils.write(aWriter, "spaceHash",contextHash);
		}
		XMLUtils.write(aWriter, "space",contextToString);
		
		XMLUtils.writeEscaped(aWriter, "endpoint",endpointToString);	
		if (physicalEndpointHash!=null)
		{	XMLUtils.writeEscaped(aWriter, "endpointHash",physicalEndpointHash);
		}
		
		IIdentifier endpointId=aEndpointId;	
		if (endpointId!=null)
		{	String logicalEndpointToString=endpointId.toString();
			if (contextHash!=null)
			{	String logicalEndpointHash = contextHash+"/"+endpointId.toString();
				XMLUtils.writeEscaped(aWriter, "logicalEndpointHash",logicalEndpointHash);
			
				IEndpointMeta meta = (IEndpointMeta)((aDMR==null)?null:aDMR.metadataForHash(logicalEndpointHash));
				if (meta!=null)
				{	String name=meta.getName();
					if (name!=null)
					{	logicalEndpointToString=name;
					}
				}
			}
			XMLUtils.writeEscaped(aWriter, "logicalEndpoint",logicalEndpointToString);
		}		
		
		int skip=0;
		Throwable t=aCause;
		
		while (t!=null && !(t instanceof RequestFrameException || t instanceof RequestFrameError))
		{	writeFrame(t, aWriter, aDMR);
			skip++;
			t=t.getCause();
		}
		
		return skip;		
	}
	
	
}
